# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""The Store API handling."""
import os
import pathlib
import platform
import time
from collections.abc import Callable
from functools import wraps
from typing import Any

import craft_store
from craft_cli import CraftError, emit
from craft_store import attenuations, endpoints
from craft_store.errors import CredentialsAlreadyAvailable
from craft_store.models.resource_revision_model import (
    CharmResourceRevision,
)
from dateutil import parser

from charmcraft import const
from charmcraft.store.client import (
    AnonymousClient,
    Client,
)
from charmcraft.store.models import (
    Account,
    Base,
    Channel,
    Entity,
    Error,
    Library,
    MacaroonInfo,
    Package,
    RegistryCredentials,
    Release,
    Resource,
    Revision,
    Uploaded,
)

# helpers to build responses from this layer


# XXX Facundo 2020-07-23: Need to do a massive rename to call `revno` to the "revision as
# the number" inside the "revision as the structure", this gets super confusing in the code with
# time, and now it's the moment to do it (also in Release below!)


# those statuses after upload that flag that the review ended (and if it ended successfully or not)
UPLOAD_ENDING_STATUSES = {
    "approved": True,
    "rejected": False,
}
POLL_DELAYS = [0.2, 0.3, 0.5, 1, 1, 1, 2] + [3] * 18  # total is 60 seconds

# default restrictions to get auth credentials
AUTH_DEFAULT_TTL = 3600 * 30
AUTH_DEFAULT_PERMISSIONS = [
    attenuations.ACCOUNT_REGISTER_PACKAGE,
    attenuations.ACCOUNT_VIEW_PACKAGES,
    attenuations.PACKAGE_MANAGE,
    attenuations.PACKAGE_VIEW,
]


def _build_errors(item):
    """Build a list of Errors from response item."""
    return [Error(message=e["message"], code=e["code"]) for e in (item["errors"] or [])]


def _build_revision(item: dict[str, Any]) -> Revision:
    """Build a Revision from a response item."""
    bases = [(None if base is None else Base(**base)) for base in item["bases"]]
    return Revision(
        revision=item["revision"],
        version=item["version"],
        created_at=parser.parse(item["created-at"]),
        status=item["status"],
        errors=_build_errors(item),
        bases=bases,
    )


def _build_library(resp):
    """Build a Library from a response."""
    return Library(
        api=resp["api"],
        content=resp.get("content"),  # not always present
        content_hash=resp["hash"],
        lib_id=resp["library-id"],
        lib_name=resp["library-name"],
        charm_name=resp["charm-name"],
        patch=resp["patch"],
    )


def _build_resource(item):
    """Build a Resource from a response item."""
    return Resource(
        name=item["name"],
        optional=item.get("optional"),
        revision=item.get("revision"),
        resource_type=item["type"],
    )


def _get_hostname() -> str:
    """Return the computer's network name or UNNKOWN if it cannot be determined."""
    hostname = platform.node()
    if not hostname:
        hostname = "UNKNOWN"
    return hostname


def _store_client_wrapper(auto_login: bool = True) -> Callable[[Callable], Callable]:
    """Decorate method to handle store error and login scenarios."""

    def store_client_wrapper_decorator(method):
        """Decorate methods to handle store errors."""

        @wraps(method)
        def error_decorator(self, *args, **kwargs):
            """Handle craft-store error situations and login scenarios."""
            try:
                return method(self, *args, **kwargs)
            except craft_store.errors.CredentialsUnavailable:
                if os.getenv(const.ALTERNATE_AUTH_ENV_VAR):
                    raise RuntimeError(
                        "Charmcraft error: internal inconsistency detected "
                        "(CredentialsUnavailable error while having user provided credentials)."
                    )
                if not auto_login:
                    raise
                emit.progress("Credentials not found. Trying to log in...")
            except craft_store.errors.StoreServerError as error:
                if error.response.status_code == 401:
                    if os.getenv(const.ALTERNATE_AUTH_ENV_VAR):
                        raise CraftError(
                            "Provided credentials are no longer valid for Charmhub. "
                            "Regenerate them and try again."
                        )
                    if not auto_login:
                        raise CraftError("Existing credentials are no longer valid for Charmhub.")
                    emit.progress("Existing credentials no longer valid. Trying to log in...")
                    # Clear credentials before trying to login again
                    self.logout()
                else:
                    raise CraftError(str(error)) from error

            self.login()

            return method(self, *args, **kwargs)

        return error_decorator

    return store_client_wrapper_decorator


class Store:
    """The main interface to the Store's API."""

    def __init__(self, charmhub_config, ephemeral=False, needs_auth=True):
        if needs_auth:
            try:
                self._client = Client(
                    charmhub_config.api_url, charmhub_config.storage_url, ephemeral=ephemeral
                )
            except craft_store.errors.NoKeyringError as error:
                raise CraftError(str(error)) from error
        else:
            self._client = AnonymousClient(charmhub_config.api_url, charmhub_config.storage_url)

    def login(self, permissions=None, ttl=None, charms=None, bundles=None, channels=None):
        """Login into the store."""
        hostname = _get_hostname()
        # Used to identify the login on Ubuntu SSO to ease future revokations.
        description = f"charmcraft@{hostname}"

        ttl = AUTH_DEFAULT_TTL if ttl is None else ttl
        permissions = AUTH_DEFAULT_PERMISSIONS if permissions is None else permissions
        kwargs = {"description": description, "ttl": ttl, "permissions": permissions}

        if channels is not None:
            kwargs["channels"] = channels

        packages = []
        if charms is not None:
            packages.extend(
                endpoints.Package(package_type="charm", package_name=charm) for charm in charms
            )
        if bundles is not None:
            packages.extend(
                endpoints.Package(package_type="bundle", package_name=bundle) for bundle in bundles
            )
        if packages:
            kwargs["packages"] = packages

        try:
            credentials = self._client.login(**kwargs)
        except CredentialsAlreadyAvailable as exc:
            raise CraftError(
                "Cannot login because credentials were found in your system "
                "(which may be no longer valid, though).",
                resolution="Please logout first, then login again.",
            ) from exc
        return credentials

    def logout(self):
        """Logout from the store.

        There's no action really in the Store to logout, we just remove local credentials.
        """
        self._client.logout()

    @_store_client_wrapper(auto_login=False)
    def whoami(self):
        """Return authenticated user details."""
        response = self._client.whoami()

        acc = response["account"]
        account = Account(name=acc["display-name"], username=acc["username"], id=acc["id"])
        if response["packages"] is None:
            packages = None
        else:
            packages = [
                Package(type=pkg["type"], name=pkg.get("name"), id=pkg.get("id"))
                for pkg in response["packages"]
            ]
        return MacaroonInfo(
            account=account,
            packages=packages,
            channels=response["channels"],
            permissions=response["permissions"],
        )

    def _check_authorized(self) -> None:
        """Check if current credentials authenticated."""
        self._client.whoami()

    @_store_client_wrapper()
    def register_name(self, name, entity_type):
        """Register the specified name for the authenticated user."""
        self._client.request_urlpath_json(
            "POST", "/v1/charm", json={"name": name, "type": entity_type}
        )

    @_store_client_wrapper()
    def unregister_name(self, name: str) -> None:
        """Unregister a package that hasn't been published.

        :param name: the package name to unregister.

        :raises: error messages from craft-store
        """
        self._client.unregister_name(name)

    @_store_client_wrapper()
    def list_registered_names(self, include_collaborations: bool) -> list[Entity]:
        """Return names registered by the authenticated user."""
        endpoint = "/v1/charm"
        if include_collaborations:
            endpoint += "?include-collaborations=true"
        response = self._client.request_urlpath_json("GET", endpoint)
        result = []
        for item in response["results"]:
            result.append(
                Entity(
                    name=item["name"],
                    private=item["private"],
                    status=item["status"],
                    entity_type=item["type"],
                    publisher_display_name=item["publisher"]["display-name"],
                )
            )
        return result

    def _upload(self, endpoint, filepath, *, extra_fields=None):
        """Upload for all charms, bundles and resources (generic process)."""
        self._check_authorized()

        upload_id = self._client.push_file(filepath)
        payload = {"upload-id": upload_id}
        if extra_fields is not None:
            payload.update(extra_fields)
        response = self._client.request_urlpath_json("POST", endpoint, json=payload)
        status_url = response["status-url"]
        emit.progress(f"Upload {upload_id} started, got status url {status_url}")

        for poll_delay in POLL_DELAYS:
            response = self._client.request_urlpath_json("GET", status_url)
            emit.progress(f"Status checked: {response}")

            # as we're asking for a single upload_id, the response will always have only one item
            (revision,) = response["revisions"]
            status = revision["status"]

            if status in UPLOAD_ENDING_STATUSES:
                return Uploaded(
                    ok=UPLOAD_ENDING_STATUSES[status],
                    errors=_build_errors(revision),
                    status=status,
                    revision=revision["revision"],
                )

            time.sleep(poll_delay)

        # no success in all the attempts to get an ending status
        raise CraftError(
            f"Timeout polling Charmhub for upload status (after {sum(POLL_DELAYS)}s)."
        )

    @_store_client_wrapper()
    def upload(self, name, filepath):
        """Upload the content of filepath to the indicated charm."""
        endpoint = f"/v1/charm/{name}/revisions"
        return self._upload(endpoint, filepath)

    @_store_client_wrapper()
    def upload_resource(
        self,
        charm_name: str,
        resource_name: str,
        resource_type,
        filepath: pathlib.Path,
        bases: list[dict[str, Any]] | None = None,
    ):
        """Upload the content of filepath to the indicated resource."""
        if bases is None:
            bases = [{"architectures": ["all"]}]
        extra_fields = {
            "type": resource_type,
            "bases": bases,
        }
        endpoint = f"/v1/charm/{charm_name}/resources/{resource_name}/revisions"
        return self._upload(endpoint, filepath, extra_fields=extra_fields)

    @_store_client_wrapper()
    def list_revisions(self, name):
        """Return charm revisions for the indicated charm."""
        response = self._client.request_urlpath_json("GET", f"/v1/charm/{name}/revisions")
        return [_build_revision(item) for item in response["revisions"]]

    @_store_client_wrapper()
    def release(self, name: str, revision: int, channels: list[str], resources) -> dict[str, Any]:
        """Release one or more revisions for a package."""
        endpoint = f"/v1/charm/{name}/releases"
        resources = [{"name": res.name, "revision": res.revision} for res in resources]
        items = [
            {"revision": revision, "channel": channel, "resources": resources}
            for channel in channels
        ]

        return self._client.request_urlpath_json("POST", endpoint, json=items)

    @_store_client_wrapper()
    def list_releases(self, name: str) -> tuple[list[Release], list[Channel], list[Revision]]:
        """List current releases for a package."""
        endpoint = f"/v1/charm/{name}/releases"
        response = self._client.request_urlpath_json("GET", endpoint)

        channel_map = []
        for item in response["channel-map"]:
            expires_at = item["expiration-date"]
            if expires_at is not None:
                # `datetime.datetime.fromisoformat` is available only since Py3.7
                expires_at = parser.parse(expires_at)
            resources = [_build_resource(r) for r in item["resources"]]
            base = None if item["base"] is None else Base(**item["base"])
            channel_map.append(
                Release(
                    revision=item["revision"],
                    channel=item["channel"],
                    expires_at=expires_at,
                    resources=resources,
                    base=base,
                )
            )

        channels = [
            Channel(
                name=item["name"],
                fallback=item["fallback"],
                track=item["track"],
                risk=item["risk"],
                branch=item["branch"],
            )
            for item in response["package"]["channels"]
        ]

        revisions = [_build_revision(item) for item in response["revisions"]]

        return channel_map, channels, revisions

    @_store_client_wrapper()
    def create_library_id(self, charm_name, lib_name):
        """Create a new library id."""
        endpoint = f"/v1/charm/libraries/{charm_name}"
        response = self._client.request_urlpath_json(
            "POST", endpoint, json={"library-name": lib_name}
        )
        return response["library-id"]

    @_store_client_wrapper()
    def create_library_revision(self, charm_name, lib_id, api, patch, content, content_hash):
        """Create a new library revision."""
        endpoint = f"/v1/charm/libraries/{charm_name}/{lib_id}"
        payload = {
            "api": api,
            "patch": patch,
            "content": content,
            "hash": content_hash,
        }
        response = self._client.request_urlpath_json("POST", endpoint, json=payload)
        return _build_library(response)

    @_store_client_wrapper()
    def get_library(self, charm_name, lib_id, api):
        """Get the library tip by id for a given api version."""
        endpoint = f"/v1/charm/libraries/{charm_name}/{lib_id}?api={api}"
        response = self._client.request_urlpath_json("GET", endpoint)
        return _build_library(response)

    @_store_client_wrapper()
    def get_libraries_tips(self, libraries):
        """Get the tip details for several libraries at once.

        Each requested library can be specified in different ways: using the library id
        or the charm and library names (both will pinpoint the library), but in the later
        case the library name is optional (so all libraries for that charm will be
        returned). Also, for all those cases, an API version can be specified.
        """
        endpoint = "/v1/charm/libraries/bulk"
        payload = []
        for lib in libraries:
            if "lib_id" in lib:
                item = {
                    "library-id": lib["lib_id"],
                }
            else:
                item = {
                    "charm-name": lib["charm_name"],
                }
                if "lib_name" in lib:
                    item["library-name"] = lib["lib_name"]
            if "api" in lib:
                item["api"] = lib["api"]
            payload.append(item)
        response = self._client.request_urlpath_json("POST", endpoint, json=payload)
        libraries = response["libraries"]
        return {(item["library-id"], item["api"]): _build_library(item) for item in libraries}

    @_store_client_wrapper()
    def list_resources(self, charm):
        """Return resources associated to the indicated charm."""
        response = self._client.request_urlpath_json("GET", f"/v1/charm/{charm}/resources")
        return [_build_resource(item) for item in response["resources"]]

    @_store_client_wrapper()
    def list_resource_revisions(
        self, charm_name: str, resource_name: str
    ) -> list[CharmResourceRevision]:
        """Return revisions for the indicated charm resource."""
        return self._client.list_resource_revisions(charm_name, resource_name)

    @_store_client_wrapper()
    def get_oci_registry_credentials(
        self, charm_name: str, resource_name: str
    ) -> RegistryCredentials:
        """Get credentials to upload a resource to the Canonical's OCI Registry."""
        endpoint = f"/v1/charm/{charm_name}/resources/{resource_name}/oci-image/upload-credentials"
        response = self._client.request_urlpath_json("GET", endpoint)
        return RegistryCredentials(
            image_name=response["image-name"],
            username=response["username"],
            password=response["password"],
        )

    @_store_client_wrapper()
    def get_oci_image_blob(self, charm_name, resource_name, digest):
        """Get the blob that points to the OCI image in the Canonical's OCI Registry."""
        payload = {"image-digest": digest}
        endpoint = f"/v1/charm/{charm_name}/resources/{resource_name}/oci-image/blob"
        return self._client.request_urlpath_text("POST", endpoint, json=payload)
        # the response here is returned as is, because it's opaque to charmcraft



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for project pre-processing functions."""
import pathlib
import textwrap

import pytest

from charmcraft import const, errors, preprocess

BASIC_BUNDLE = {"type": "bundle", "parts": {"bundle": {"plugin": "bundle", "source": "."}}}
BASIC_CHARM = {"type": "charm", "parts": {"charm": {"plugin": "charm", "source": "."}}}
BASIC_BASES_CHARM = {**BASIC_CHARM, "bases": [{"name": "ubuntu", "channel": "22.04"}]}


@pytest.mark.parametrize(
    ("yaml_data", "expected"),
    [
        pytest.param({}, {}, id="no-type"),
        pytest.param({"type": "bundle"}, BASIC_BUNDLE, id="empty-bundle"),
        pytest.param(BASIC_BUNDLE.copy(), BASIC_BUNDLE, id="prefilled-bundle"),
        pytest.param(
            {"type": "charm", "bases": []},
            {"type": "charm", "bases": [], "parts": {"charm": {"plugin": "charm", "source": "."}}},
            id="empty-charm",
        ),
        pytest.param(BASIC_CHARM.copy(), BASIC_CHARM, id="basic-charm"),
    ],
)
def test_add_default_parts_correct(yaml_data, expected):
    preprocess.add_default_parts(yaml_data)

    assert yaml_data == expected


@pytest.mark.parametrize(
    ("yaml_data", "metadata_yaml", "expected"),
    [
        pytest.param({}, None, {}, id="nonexistent"),
        pytest.param({}, "{}", {"name": None, "summary": None, "description": None}, id="empty"),
        pytest.param(
            {"name": "my-charm"},
            "summary: a charm",
            {"name": "my-charm", "summary": "a charm", "description": None},
            id="merge",
        ),
        pytest.param(
            {},
            "name: my-charm",
            {"name": "my-charm", "summary": None, "description": None},
            id="only-from-metadata",
        ),
    ],
)
def test_add_metadata_success(fs, yaml_data, metadata_yaml, expected):
    if metadata_yaml is not None:
        fs.create_file("/project/metadata.yaml", contents=metadata_yaml)

    preprocess.add_metadata(pathlib.Path("/project"), yaml_data)

    assert yaml_data == expected


@pytest.mark.parametrize(
    ("yaml_data", "metadata_yaml", "message"),
    [
        (
            {
                "name": "test-charm",
                "summary": "A test charm",
                "description": "A charm for testing!",
            },
            "",
            "Invalid file: 'metadata.yaml'",
        ),
        (
            {
                "name": "test-charm",
                "summary": "A test charm",
                "description": "A charm for testing!",
            },
            textwrap.dedent(
                """\
                name: test-charm
                summary: A test charm
                description: A charm for testing!"""
            ),
            "Fields in charmcraft.yaml cannot be duplicated in metadata.yaml",
        ),
    ],
)
def test_extra_yaml_transform_failure(fs, yaml_data, metadata_yaml, message):
    fs.create_file("metadata.yaml", contents=metadata_yaml)

    with pytest.raises(errors.CraftError) as exc_info:
        preprocess.add_metadata(pathlib.Path("/"), yaml_data)

    assert exc_info.value.args[0] == message


@pytest.mark.parametrize(
    ("yaml_data", "bundle_yaml", "expected"),
    [
        pytest.param({}, "", {}, id="non-bundle"),
        pytest.param(
            {"type": "bundle"}, "{}", {"type": "bundle", "bundle": {}}, id="empty-bundle"
        ),
    ],
)
def test_add_bundle_snippet_success(fs, yaml_data, bundle_yaml, expected):
    project_dir = pathlib.Path("project")
    fs.create_file(project_dir / const.BUNDLE_FILENAME, contents=bundle_yaml)

    preprocess.add_bundle_snippet(project_dir, yaml_data)

    assert yaml_data == expected


def test_add_bundle_snippet_no_file(fs):
    fs.create_dir("project")

    with pytest.raises(
        errors.CraftError, match="^Missing 'bundle.yaml' file: '.*bundle.yaml'"
    ) as exc_info:
        preprocess.add_bundle_snippet(pathlib.Path("/project"), {"type": "bundle"})

    assert exc_info.value.retcode == 66


@pytest.mark.parametrize("contents", ["", "A string", "0", "[]"])
def test_add_bundle_snippet_invalid_file(fs, contents):
    fs.create_file("project/bundle.yaml", contents=contents)

    with pytest.raises(
        errors.CraftError, match="Incorrectly formatted 'bundle.yaml' file"
    ) as exc_info:
        preprocess.add_bundle_snippet(pathlib.Path("project"), {"type": "bundle"})

    assert exc_info.value.retcode == 65


@pytest.mark.parametrize(
    ("yaml_data", "config_yaml", "expected"),
    [
        ({}, "{}", {"config": {}}),
        ({}, "options:\n boop:\n  type: int", {"config": {"options": {"boop": {"type": "int"}}}}),
    ],
)
def test_add_config_success(fs, yaml_data, config_yaml, expected):
    project_dir = pathlib.Path("project")
    fs.create_file(project_dir / const.JUJU_CONFIG_FILENAME, contents=config_yaml)

    preprocess.add_config(project_dir, yaml_data)

    assert yaml_data == expected


@pytest.mark.parametrize(
    ("yaml_data", "actions_yaml", "expected"),
    [
        ({}, "", {}),
        ({}, "{}", {}),
        (
            {},
            "boop:\n description: Boop in the snoot",
            {"actions": {"boop": {"description": "Boop in the snoot"}}},
        ),
    ],
)
def test_add_actions_success(fs, yaml_data, actions_yaml, expected):
    project_dir = pathlib.Path("project")
    fs.create_file(project_dir / const.JUJU_ACTIONS_FILENAME, contents=actions_yaml)

    preprocess.add_actions(project_dir, yaml_data)

    assert yaml_data == expected



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Extension base class definition."""

import abc
import os
import sys
from collections.abc import Sequence
from pathlib import Path
from typing import Any, Final, final

from craft_cli import emit

from charmcraft import const, errors


class Extension(abc.ABC):
    """Extension is the class from which all extensions inherit.

    Extensions have the ability to add snippets to charms, parts, config, actions,
    and indeed add new parts to a given charmcraft.yaml.
    :ivar project_root: the root of the project.
    :ivar yaml_data: the raw yaml data.
    """

    def __init__(
        self,
        *,
        project_root: Path,
        yaml_data: dict[str, Any],
    ) -> None:
        """Create a new Extension."""
        self.project_root = project_root
        self.yaml_data: Final[dict[str, Any]] = yaml_data

    @staticmethod
    @abc.abstractmethod
    def get_supported_bases() -> list[tuple[str, str]]:
        """Return a list of tuple of supported bases."""

    @staticmethod
    @abc.abstractmethod
    def is_experimental(base: tuple[str, str] | None) -> bool:
        """Return whether or not this extension is unstable for given base."""

    @abc.abstractmethod
    def get_root_snippet(self) -> dict[str, Any]:
        """Return the root snippet to apply."""

    @abc.abstractmethod
    def get_part_snippet(self) -> dict[str, Any]:
        """Return the part snippet to apply to existing parts."""

    @abc.abstractmethod
    def get_parts_snippet(self) -> dict[str, Any]:
        """Return the parts to add to parts."""

    @final
    def validate(self, extension_name: str):
        """Validate that the extension can be used with the current project.

        :param extension_name: the name of the extension being parsed.
        :raises errors.ExtensionError: if the extension is incompatible with the project.
        """
        if "bases" not in self.yaml_data:
            # There is nothing to validate, the extension will set the preferred base.
            return

        bases: list = self.yaml_data["bases"]

        for base in bases:
            for build_on in base.get("build-on", []):
                build_base = (build_on["name"], build_on["channel"])

                if self.is_experimental(build_base) and not os.getenv(
                    const.EXPERIMENTAL_EXTENSIONS_ENV_VAR
                ):
                    raise errors.ExtensionError(
                        f"Extension is experimental: {extension_name!r}",
                        docs_url="https://juju.is/docs/sdk/charmcraft-config",  # no docs yet
                    )

                if self.is_experimental(build_base):
                    emit.progress(
                        f"*EXPERIMENTAL* extension {extension_name!r} enabled",
                        permanent=True,
                    )

                if build_base not in self.get_supported_bases():
                    raise errors.ExtensionError(
                        f"Extension {extension_name!r} does not support base: {build_base!r}"
                    )

        invalid_parts = [
            p for p in self.get_parts_snippet() if not p.startswith(f"{extension_name}/")
        ]
        if invalid_parts:
            raise ValueError(
                f"Extension has invalid part names: {invalid_parts!r}. "
                "Format is <extension-name>/<part-name>"
            )


def get_extensions_data_dir() -> Path:
    """Return the path to the extension data directory."""
    return Path(sys.prefix) / "share" / "charmcraft" / "extensions"


def append_to_env(env_variable: str, paths: Sequence[str], separator: str = ":") -> str:
    """Return a string for env_variable with one of more paths appended.

    :param env_variable: the variable to operate on.
    :param paths: one or more paths to append.
    :param separator: the separator to use.
    :returns: a shell string where one or more paths are appended
                  to env_variable. The code takes into account the case
                  where the environment variable is empty, to avoid putting
                  a separator token at the start.
    """
    return f"${{{env_variable}:+${env_variable}{separator}}}" + separator.join(paths)


def prepend_to_env(env_variable: str, paths: Sequence[str], separator: str = ":") -> str:
    """Return a string for env_variable with one of more paths prepended.

    :param env_variable: the variable to operate on.
    :param paths: one or more paths to append.
    :param separator: the separator to use.
    :returns: a shell string where one or more paths are prepended
                  before env_variable. The code takes into account the case
                  where the environment variable is empty, to avoid putting
                  a separator token at the end.
    """
    return separator.join(paths) + f"${{{env_variable}:+{separator}${env_variable}}}"



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for store models."""

import pytest

from charmcraft.store import models


@pytest.mark.parametrize(
    ("data", "expected"),
    [
        (
            {
                "charm-name": "abc",
                "library-name": "def",
                "library-id": "id",
                "api": 1,
                "patch": 123,
                "hash": "hashyhash",
            },
            models.Library(
                charm_name="abc",
                lib_name="def",
                lib_id="id",
                api=1,
                patch=123,
                content_hash="hashyhash",
                content=None,
            ),
        ),
        (
            {
                "charm-name": "abc",
                "library-name": "def",
                "library-id": "id",
                "api": 1,
                "patch": 123,
                "hash": "hashyhash",
                "content": "I am a library.",
            },
            models.Library(
                charm_name="abc",
                lib_name="def",
                lib_id="id",
                api=1,
                patch=123,
                content_hash="hashyhash",
                content="I am a library.",
            ),
        ),
    ],
)
def test_library_from_dict(data, expected):
    assert models.Library.from_dict(data) == expected



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Charmcraft-specific overrides for the remote build service."""
import datetime
import pathlib
from collections.abc import Mapping
from typing import Any

import lazr.restfulclient.errors  # type: ignore[import-untyped]
from craft_application import launchpad
from craft_application.services import remotebuild
from overrides import override


class RemoteBuildService(remotebuild.RemoteBuildService):
    """Charmcraft remote build service."""

    RecipeClass = launchpad.models.CharmRecipe

    def fetch_logs(self, output_dir: pathlib.Path) -> Mapping[str, pathlib.Path | None]:
        """Fetch the logs for each build to the given directory.

        :param output_dir: The directory into which to place the logs.
        :returns: A mapping of the architecture to its build log.
        """
        if not self._is_setup:
            raise RuntimeError(
                "RemoteBuildService must be set up using start_builds or resume_builds before fetching logs."
            )
        project_name = self._name.split("-", maxsplit=2)[1]
        logs: dict[str, pathlib.Path | None] = {}
        log_downloads: dict[str, pathlib.Path] = {}
        fetch_time = datetime.datetime.now().isoformat(timespec="seconds")
        for build in self._builds:
            url = build.build_log_url
            if not url:
                logs[build.arch_tag] = None
                continue
            filename = f"{project_name}_{build.distribution.name}-{build.distro_series.version}-{build.arch_tag}-{fetch_time}.txt"
            logs[build.arch_tag] = output_dir / filename
            log_downloads[url] = output_dir / filename
        self.request.download_files_with_progress(log_downloads)
        return logs

    @override
    def _new_recipe(
        self,
        name: str,
        repository: launchpad.models.GitRepository,
        **_: Any,  # noqa: ANN401
    ) -> launchpad.models.Recipe:
        """Create a new recipe."""
        try:
            return launchpad.models.CharmRecipe.new(
                self.lp,
                name,
                self.lp.username,
                project=self._lp_project.name,
                git_ref=repository.self_link + "/+ref/main",
            )
        except lazr.restfulclient.errors.BadRequest:
            return self.lp.get_recipe(
                "CHARM",
                name=name,
                owner=self.lp.username,
                project=self._lp_project.name,
            )

    @override
    def _get_build_states(self) -> Mapping[str, launchpad.models.BuildState]:
        self._refresh_builds()
        return {
            f"{build.distribution.name}@{build.distro_series.version} ({build.arch_tag})": build.get_state()
            for build in self._builds
        }



# Copyright 2023-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for package service."""

import datetime
import pathlib
import sys
import zipfile
from typing import Any

import craft_cli.pytest_plugin
import pytest
import pytest_check
from craft_application import util
from craft_application.models import BuildInfo
from craft_providers.bases import BaseName

from charmcraft import const, models, services
from charmcraft.application.main import APP_METADATA
from charmcraft.models.project import BasesCharm

SIMPLE_BUILD_BASE = models.charmcraft.Base(name="ubuntu", channel="22.04", architectures=["arm64"])
SIMPLE_MANIFEST = models.Manifest(
    charmcraft_started_at="1970-01-01T00:00:00+00:00",
    bases=[SIMPLE_BUILD_BASE],
)
MANIFEST_WITH_ATTRIBUTE = models.Manifest(
    **SIMPLE_MANIFEST.marshal(),
    analysis={"attributes": [models.Attribute(name="boop", result="success")]},
)


@pytest.fixture
def package_service(fake_path, simple_charm, service_factory, default_build_plan):
    fake_project_dir = fake_path / "project"
    fake_project_dir.mkdir(parents=True)

    service_factory.set_kwargs(
        "lifecycle",
        work_dir=fake_path,
        cache_dir=fake_path / "cache",
        build_plan=[],  # Only okay now because we're not asking the lifecycle service to use the plan.
    )

    return services.PackageService(
        app=APP_METADATA,
        project=simple_charm,
        # The package service doesn't call other services
        services=service_factory,
        project_dir=fake_project_dir,
        build_plan=default_build_plan,
    )


@pytest.mark.parametrize(
    "metadata",
    [
        models.CharmMetadata(
            name="charmy-mccharmface",
            summary="Charmy!",
            description="Very charming!",
        ),
    ],
)
def test_get_metadata(package_service, simple_charm: BasesCharm, metadata):
    package_service._project = simple_charm

    assert package_service.metadata == metadata


@pytest.mark.parametrize(
    ("bases", "expected_name"),
    [
        (
            [SIMPLE_BUILD_BASE],
            "charmy-mccharmface_distro-1-test64.charm",
        ),
    ],
)
def test_get_charm_path(fake_path, package_service, bases, expected_name):
    fake_prime_dir = fake_path / "prime"
    charm_path = package_service.get_charm_path(fake_prime_dir)

    assert charm_path == fake_prime_dir / expected_name


@pytest.mark.parametrize(
    ("lint", "expected"),
    [
        ([], SIMPLE_MANIFEST),
        (
            [models.CheckResult("lint", "lint", "lint", models.CheckType.LINT, "")],
            SIMPLE_MANIFEST,
        ),
        (
            [models.CheckResult("boop", "success", "", models.CheckType.ATTRIBUTE, "")],
            MANIFEST_WITH_ATTRIBUTE,
        ),
    ],
)
def test_get_manifest(package_service, simple_charm, lint, expected):
    simple_charm._started_at = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)

    assert package_service.get_manifest(lint) == expected


def test_do_not_overwrite_metadata_yaml(
    emitter: craft_cli.pytest_plugin.RecordingEmitter,
    fake_path,
    package_service,
    simple_charm,
):
    fake_prime_dir = fake_path / "prime"
    fake_prime_dir.mkdir()
    fake_stage_dir = fake_path / "stage"
    fake_stage_dir.mkdir()
    fake_staged_metadata = fake_stage_dir / const.METADATA_FILENAME
    fake_staged_metadata.touch()
    package_service._project.parts["reactive"] = {"source": "."}

    package_service.write_metadata(fake_prime_dir)

    emitter.assert_debug(
        "'metadata.yaml' generated by charm. Not using original project metadata."
    )


def test_do_not_overwrite_actions_yaml(
    emitter: craft_cli.pytest_plugin.RecordingEmitter,
    fake_path,
    package_service,
    simple_charm,
):
    fake_prime_dir = fake_path / "prime"
    fake_prime_dir.mkdir()
    fake_stage_dir = fake_path / "stage"
    fake_stage_dir.mkdir()
    fake_staged_actions = fake_stage_dir / const.JUJU_ACTIONS_FILENAME
    fake_staged_actions.touch()
    package_service._project.parts["reactive"] = {"source": "."}

    package_service.write_metadata(fake_prime_dir)

    emitter.assert_debug("'actions.yaml' generated by charm. Skipping generation.")


# region Tests for getting bases for manifest.yaml
@pytest.mark.parametrize(
    ("bases", "build_item", "expected"),
    [
        (
            [{"name": "ubuntu", "channel": "20.04"}],
            BuildInfo(
                platform=util.get_host_architecture(),
                build_for=util.get_host_architecture(),
                build_on=util.get_host_architecture(),
                base=BaseName("ubuntu", "20.04"),
            ),
            [
                {
                    "name": "ubuntu",
                    "channel": "20.04",
                    "architectures": [util.get_host_architecture()],
                }
            ],
        ),
        (
            [
                {
                    "build-on": [
                        {"name": "ubuntu", "channel": "22.04", "architectures": ["riscv64"]}
                    ],
                    "run-on": [
                        {"name": "ubuntu", "channel": "22.04", "architectures": ["all"]},
                    ],
                },
            ],
            BuildInfo(
                platform="riscv64",
                build_for="riscv64",
                build_on="riscv64",
                base=BaseName("ubuntu", "22.04"),
            ),
            [
                {"name": "ubuntu", "channel": "22.04", "architectures": ["all"]},
            ],
        ),
        (
            [{"name": "centos", "channel": "7"}],
            BuildInfo(
                platform=util.get_host_architecture(),
                build_on=util.get_host_architecture(),
                build_for=util.get_host_architecture(),
                base=BaseName("centos", "7"),
            ),
            [{"name": "centos", "channel": "7", "architectures": [util.get_host_architecture()]}],
        ),
        pytest.param(
            [
                {"name": "centos", "channel": "7"},
                {
                    "build-on": [{"name": "ubuntu", "channel": "20.04"}],
                    "run-on": [{"name": "ubuntu", "channel": "20.04", "architectures": ["all"]}],
                },
                {
                    "build-on": [
                        {"name": "ubuntu", "channel": "22.04", "architectures": ["amd64", "arm64"]}
                    ],
                    "run-on": [{"name": "ubuntu", "channel": "22.04", "architectures": ["arm64"]}],
                },
            ],
            BuildInfo(
                platform="amd64",
                build_on="amd64",
                build_for="arm64",
                base=BaseName("ubuntu", "22.04"),
            ),
            [{"name": "ubuntu", "channel": "22.04", "architectures": ["arm64"]}],
            id="cross-compile",
        ),
    ],
)
def test_get_manifest_bases_from_bases(
    fake_path: pathlib.Path,
    package_service: services.PackageService,
    bases: list[dict[str, Any]],
    build_item: BuildInfo,
    expected: list[dict[str, Any]],
):
    charm = models.BasesCharm.model_validate(
        {
            "name": "my-charm",
            "description": "",
            "summary": "",
            "type": "charm",
            "bases": bases,
        }
    )
    package_service._project = charm
    package_service._build_plan = [build_item]

    assert package_service.get_manifest_bases() == [
        models.Base.model_validate(b) for b in expected
    ]


@pytest.mark.parametrize("base", ["ubuntu@22.04", "almalinux@9"])
@pytest.mark.parametrize(
    ("platforms", "selected_platform", "expected_architectures"),
    [
        ({"armhf": None}, "armhf", ["armhf"]),
        (
            {
                "anything": {
                    "build-on": [*const.SUPPORTED_ARCHITECTURES],
                    "build-for": ["all"],
                }
            },
            "anything",
            ["all"],
        ),
        (
            {
                "anything": {
                    "build-on": [*const.SUPPORTED_ARCHITECTURES],
                    "build-for": ["all"],
                },
                "amd64": None,
                "riscy": {
                    "build-on": ["arm64", "ppc64el", "riscv64"],
                    "build-for": ["all"],
                },
            },
            "anything",
            ["all"],
        ),
        ({util.get_host_architecture(): None}, None, [util.get_host_architecture()]),
    ],
)
def test_get_manifest_bases_from_platforms(
    package_service, base, platforms, selected_platform, expected_architectures
):
    charm = models.PlatformCharm.model_validate(
        {
            "name": "my-charm",
            "description": "",
            "summary": "",
            "type": "charm",
            "base": base,
            "platforms": platforms,
            "parts": {},
        }
    )
    package_service._project = charm
    package_service._platform = selected_platform

    bases = package_service.get_manifest_bases()

    pytest_check.equal(len(bases), 1)
    actual_base = bases[0]
    pytest_check.equal(f"{actual_base.name}@{actual_base.channel}", base)
    pytest_check.equal(actual_base.architectures, expected_architectures)


# endregion
# region tests for packing the charm
# These tests are modified from test_zipbuild
def test_pack_charm_simple(fake_path, package_service):
    """Build a bunch of files in the zip."""
    build_dir = fake_path / "somedir"
    build_dir.mkdir()

    testfile1 = build_dir / "foo.txt"
    testfile1.write_bytes(b"123\x00456")
    subdir = build_dir / "bar"
    subdir.mkdir()
    testfile2 = subdir / "baz.txt"
    testfile2.write_bytes(b"mo\xc3\xb1o")

    package_service.pack_charm(build_dir, fake_path)

    zf = zipfile.ZipFile(fake_path / "charmy-mccharmface_distro-1-test64.charm")
    assert sorted(x.filename for x in zf.infolist()) == ["bar/baz.txt", "foo.txt"]
    assert zf.read("foo.txt") == b"123\x00456"
    assert zf.read("bar/baz.txt") == b"mo\xc3\xb1o"


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_zipbuild_symlink_simple(fake_path, package_service):
    """Symlinks are supported."""
    build_dir = fake_path / "somedir"
    build_dir.mkdir()

    testfile1 = build_dir / "real.txt"
    testfile1.write_bytes(b"123\x00456")
    testfile2 = build_dir / "link.txt"
    testfile2.symlink_to(testfile1)

    package_service.pack_charm(build_dir, fake_path)

    zf = zipfile.ZipFile(fake_path / "charmy-mccharmface_distro-1-test64.charm")
    assert sorted(x.filename for x in zf.infolist()) == ["link.txt", "real.txt"]
    assert zf.read("real.txt") == b"123\x00456"
    assert zf.read("link.txt") == b"123\x00456"


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_zipbuild_symlink_outside(fake_path, package_service):
    """No matter where the symlink points to."""
    # outside the build dir
    testfile1 = fake_path / "real.txt"
    testfile1.write_bytes(b"123\x00456")

    # inside the build dir
    build_dir = fake_path / "somedir"
    build_dir.mkdir()
    testfile2 = build_dir / "link.txt"
    testfile2.symlink_to(testfile1)

    package_service.pack_charm(build_dir, fake_path)

    zf = zipfile.ZipFile(fake_path / "charmy-mccharmface_distro-1-test64.charm")
    assert sorted(x.filename for x in zf.infolist()) == ["link.txt"]
    assert zf.read("link.txt") == b"123\x00456"


# endregion



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for the lifecycle service."""


import distro
import pytest
from craft_application import errors, models, util
from craft_providers import bases

from charmcraft.services import CharmcraftServiceFactory


def test_init_lifecycle(service_factory: CharmcraftServiceFactory):
    """Test the setup of a parts lifecycle, implicitly testing setup."""

    service_factory.lifecycle._init_lifecycle_manager()


def test_lifecycle_build_for_invalid(service_factory: CharmcraftServiceFactory):
    lifecycle = service_factory.lifecycle

    lifecycle._build_plan = [
        models.BuildInfo(
            platform="something",
            build_on=util.get_host_architecture(),
            build_for="invalid",
            base=bases.BaseName(distro.id(), distro.version()),
        )
    ]

    with pytest.raises(
        errors.PartsLifecycleError, match="[Aa]rchitecture '[a-z]+' is not supported"
    ):
        lifecycle._init_lifecycle_manager()


def test_lifecycle_build_for_all(service_factory: CharmcraftServiceFactory):
    lifecycle = service_factory.lifecycle

    lifecycle._build_plan = [
        models.BuildInfo(
            platform="something",
            build_on=util.get_host_architecture(),
            build_for="all",
            base=bases.BaseName(distro.id(), distro.version()),
        )
    ]

    lcm = lifecycle._init_lifecycle_manager()

    assert lcm._target_arch == util.get_host_architecture()


def test_lifecycle_build_for_multi(service_factory: CharmcraftServiceFactory):
    lifecycle = service_factory.lifecycle

    host_arch = util.get_host_architecture()
    arches = {"arm64", "riscv64", "amd64"} - {host_arch}
    foreign_arch = next(iter(arches))

    lifecycle._build_plan = [
        models.BuildInfo(
            platform="something",
            build_on=util.get_host_architecture(),
            build_for=f"{foreign_arch}-{host_arch}",
            base=bases.BaseName(distro.id(), distro.version()),
        )
    ]

    lcm = lifecycle._init_lifecycle_manager()

    assert lcm._target_arch == foreign_arch



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Service class for store interaction."""
from __future__ import annotations

import platform
from collections.abc import Collection, Mapping, Sequence

import craft_application
import craft_store
from craft_cli import emit
from craft_store import models
from overrides import override

from charmcraft import const, env, errors, store
from charmcraft.models import CharmLib
from charmcraft.store import AUTH_DEFAULT_PERMISSIONS, AUTH_DEFAULT_TTL
from charmcraft.store.models import Library, LibraryMetadataRequest


class BaseStoreService(craft_application.AppService):
    """Business logic for interacting with the store.

    This service should be easily adjustable for craft-application.
    """

    ClientClass: type[craft_store.StoreClient] = craft_store.StoreClient
    client: craft_store.StoreClient
    _endpoints: craft_store.endpoints.Endpoints = craft_store.endpoints.CHARMHUB
    _environment_auth: str = const.ALTERNATE_AUTH_ENV_VAR

    @property
    def _user_agent(self) -> str:
        """Get the user agent string for the store service."""
        product = self._app.name
        version = self._app.version
        system_information = self._ua_system_info
        return f"{product.title()}/{version} ({system_information})"

    @property
    def _ua_system_info(self) -> str:
        """System information for user agent."""
        segments = [
            f"{platform.system()} {platform.release()}",
            platform.machine(),
            f"{platform.python_implementation()} {platform.python_version()}",
        ]
        if platform.system() == "Linux":
            import distro

            segments.append(f"{distro.name()} {distro.version()}")

        return "; ".join(segments)

    @property
    def _base_url(self) -> str:
        """Get the store base URL."""
        return env.get_store_config().api_url

    @property
    def _storage_url(self) -> str:
        return env.get_store_config().storage_url

    def setup(self) -> None:
        """Set up the store service."""
        super().setup()

        try:
            self.client = self.ClientClass(
                application_name=self._app.name,
                base_url=self._base_url,
                storage_base_url=self._storage_url,
                endpoints=self._endpoints,
                environment_auth=self._environment_auth,
                user_agent=self._user_agent,
            )
        except craft_store.errors.NoKeyringError:
            emit.progress(
                "WARNING: Cannot get a keyring. Every store interaction that requires "
                "authentication will require you to log in again.",
                permanent=True,
            )
            self.client = self.ClientClass(
                application_name=self._app.name,
                base_url=self._base_url,
                storage_base_url=self._storage_url,
                endpoints=self._endpoints,
                environment_auth=self._environment_auth,
                user_agent=self._user_agent,
                ephemeral=True,
            )

    def _get_description(self, description: str | None = None) -> str:
        """Return the given description or a default one."""
        if description is None:
            return f"{self._app.name}@{platform.node()}"
        return description

    def login(
        self,
        permissions: Sequence[str] = AUTH_DEFAULT_PERMISSIONS,
        description: str | None = None,
        ttl: int = AUTH_DEFAULT_TTL,
        packages: Sequence[craft_store.endpoints.Package] | None = None,
        channels: Sequence[str] | None = None,
    ) -> str:
        """Login to the store."""
        description = self._get_description(description)

        try:
            return self.client.login(
                permissions=permissions,
                description=description,
                ttl=ttl,
                packages=packages,
                channels=channels,
            )
        except craft_store.errors.CredentialsAlreadyAvailable as exc:
            raise errors.CraftError(
                "Cannot login because credentials were found in your system "
                "(which may be no longer valid, though).",
                resolution="Please log out first, then log in again.",
            ) from exc

    def logout(self) -> None:
        """Log out of the store."""
        self.client.logout()

    def get_account_info(self):
        """Get the account details of the logged-in account."""
        return self.client.whoami()["account"]

    def get_credentials(
        self,
        permissions: Sequence[str] = AUTH_DEFAULT_PERMISSIONS,
        description: str | None = None,
        ttl: int = AUTH_DEFAULT_TTL,
        packages: Sequence[craft_store.endpoints.Package] | None = None,
        channels: Sequence[str] | None = None,
    ) -> str:
        """Create a fresh set of login credentials for the store.

        This logs in independent of any credentials currently stored for the application and
        returns the resulting macaroon as a string.
        """
        description = self._get_description(description)

        store = craft_store.StoreClient(
            application_name=self._app.name,
            base_url=self._base_url,
            storage_base_url=self._storage_url,
            endpoints=self._endpoints,
            environment_auth=None,
            user_agent=self._user_agent,
            ephemeral=True,
        )

        return store.login(
            permissions=permissions,
            description=description,
            ttl=ttl,
            packages=packages,
            channels=channels,
        )


class StoreService(BaseStoreService):
    """A Store service specifically for Charmcraft."""

    ClientClass = store.Client
    client: store.Client  # pyright: ignore[reportIncompatibleVariableOverride]
    anonymous_client: store.AnonymousClient

    @override
    def setup(self) -> None:
        """Set up the store service."""
        super().setup()
        self.anonymous_client = store.AnonymousClient(
            api_base_url=self._base_url,
            storage_base_url=self._storage_url,
        )

    def set_resource_revisions_architectures(
        self, name: str, resource_name: str, updates: dict[int, list[str]]
    ) -> Collection[models.resource_revision_model.CharmResourceRevision]:
        """Set the architectures for one or more resource revisions.

        :param name: The name of the charm in the store
        :param resource_name: The name of the specific resource
        :param updates: A mapping of resource revision to its architectures.
        :returns: The updated revisions, as craft_store CharmResourceRevision objects.
        """
        self.client.update_resource_revisions(
            *(
                models.CharmResourceRevisionUpdateRequest(
                    revision=revision,
                    bases=[models.RequestCharmResourceBase(architectures=architectures)],
                )
                for revision, architectures in updates.items()
            ),
            name=name,
            resource_name=resource_name,
        )
        new_revisions = self.client.list_resource_revisions(name=name, resource_name=resource_name)
        return [rev for rev in new_revisions if int(rev.revision) in updates]

    def get_libraries_metadata(self, libraries: Sequence[CharmLib]) -> Sequence[Library]:
        """Get the metadata for one or more charm libraries.

        :param libraries: A sequence of libraries to request.
        :returns: A sequence of the libraries' metadata in the store.
        """
        store_libs = []
        for lib in libraries:
            charm_name, _, lib_name = lib.lib.partition(".")
            store_lib = LibraryMetadataRequest(
                {
                    "charm-name": charm_name,
                    "library-name": lib_name,
                    "api": lib.api_version,
                }
            )
            if (patch_version := lib.patch_version) is not None:
                store_lib["patch"] = patch_version
            store_libs.append(store_lib)

        return self.anonymous_client.fetch_libraries_metadata(store_libs)

    def get_libraries_metadata_by_name(
        self, libraries: Sequence[CharmLib]
    ) -> Mapping[str, Library]:
        """Get a mapping of [charm_name].[library_name] to the requested libraries."""
        return {
            f"{lib.charm_name}.{lib.lib_name}": lib
            for lib in self.get_libraries_metadata(libraries)
        }

    def get_library(
        self, charm_name: str, *, library_id: str, api: int | None = None, patch: int | None = None
    ) -> Library:
        """Get a library by charm name and ID from charmhub."""
        return self.anonymous_client.get_library(
            charm_name=charm_name, library_id=library_id, api=api, patch=patch
        )



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""YAML-related utilities for Charmcraft."""

from typing import Any

import pydantic
import yaml
from craft_cli import emit

from charmcraft import const


def load_yaml(fpath) -> dict[str, Any] | None:
    """Return the content of a YAML file."""
    if not fpath.is_file():
        emit.debug(f"Couldn't find config file {str(fpath)!r}")
        return None
    try:
        with fpath.open("r") as fh:
            content = yaml.safe_load(fh)
    except (yaml.error.YAMLError, OSError) as err:
        emit.debug(f"Failed to read/parse config file {str(fpath)!r}: {err!r}")
        return None
    return content


def _repr_str(dumper: yaml.SafeDumper, data: str) -> yaml.ScalarNode:
    """Multi-line string representer for the YAML dumper."""
    if "\n" in data:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


def dump_yaml(data: Any) -> str:  # noqa: ANN401: yaml.dump takes anything, so why can't we?
    """Dump a craft model to a YAML string."""
    yaml.add_representer(str, _repr_str, Dumper=yaml.SafeDumper)
    yaml.add_representer(
        pydantic.AnyHttpUrl, _repr_str, Dumper=yaml.SafeDumper  # type: ignore[arg-type]
    )
    yaml.add_representer(
        const.CharmArch,
        yaml.representer.SafeRepresenter.represent_str,
        Dumper=yaml.SafeDumper,
    )
    return yaml.dump(data, Dumper=yaml.SafeDumper, sort_keys=False)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for Charmcraft models."""
import pytest

from charmcraft.models import charmcraft


@pytest.mark.parametrize(
    ("base_str", "expected"),
    [
        ("ubuntu@24.04", charmcraft.Base(name="ubuntu", channel="24.04", architectures=[])),
        ("ubuntu@22.04", charmcraft.Base(name="ubuntu", channel="22.04", architectures=[])),
        ("almalinux@9", charmcraft.Base(name="almalinux", channel="9", architectures=[])),
    ],
)
def test_get_base_from_str_and_arch(base_str, expected):
    assert charmcraft.Base.from_str_and_arch(base_str, []) == expected



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for the Image service."""


import itertools
import json
from unittest import mock

import docker
import docker.errors
import docker.models.images
import pytest

from charmcraft import application, const, services, utils


@pytest.fixture
def mock_docker() -> mock.Mock:
    return mock.Mock(spec_set=docker.DockerClient)


@pytest.fixture
def mock_skopeo(fake_process) -> mock.Mock:
    fake_process.register(["/skopeo", "--version"])
    return mock.Mock(wraps=utils.Skopeo(skopeo_path="/skopeo"))


@pytest.fixture
def image_service(service_factory, mock_skopeo, mock_docker) -> services.ImageService:
    service = services.ImageService(app=application.APP_METADATA, services=service_factory)
    service._skopeo = mock_skopeo
    service._docker = mock_docker
    return service


def test_get_maybe_id_from_docker_success(image_service: services.ImageService, mock_docker):
    expected = "sha256:some-sha-hash"
    mock_docker.images.get.return_value = docker.models.images.Image(attrs={"Id": expected})

    result = image_service.get_maybe_id_from_docker("some-image")

    mock_docker.images.get.assert_called_once_with("some-image")
    assert result == expected


def test_get_maybe_id_from_docker_failure(image_service: services.ImageService, mock_docker):
    mock_docker.images.get.side_effect = docker.errors.ImageNotFound("womp womp")

    assert image_service.get_maybe_id_from_docker("some-image") is None


@pytest.mark.parametrize("image", ["my-image"])
@pytest.mark.parametrize("architecture", const.CharmArch)
def test_inspect_single_arch(
    fake_process, image_service: services.ImageService, mock_skopeo, image: str, architecture
):
    fake_process.register(
        ["/skopeo", "inspect", "--raw", image], stdout=json.dumps({"raw_manifest": True})
    )
    fake_process.register(
        ["/skopeo", "inspect", image],
        stdout=json.dumps({"Digest": "Reader's", "Architecture": architecture}),
    )

    actual = image_service.inspect(image)

    assert actual == services.image.OCIMetadata(
        path=image, digest="Reader's", architectures=[architecture]
    )


@pytest.mark.parametrize("image", ["my-image"])
@pytest.mark.parametrize("architectures", itertools.product(const.CharmArch, repeat=2))
def test_inspect_two_arch(
    fake_process, image_service: services.ImageService, mock_skopeo, image: str, architectures
):
    fake_process.register(
        ["/skopeo", "inspect", "--raw", image],
        stdout=json.dumps(
            {
                "manifests": [
                    {"platform": {"os": "linux", "architecture": str(arch)}}
                    for arch in architectures
                ],
            }
        ),
    )
    fake_process.register(
        ["/skopeo", "inspect", image],
        stdout=json.dumps({"Digest": "Reader's", "Architecture": "amd64"}),
    )

    actual = image_service.inspect(image)

    assert actual == services.image.OCIMetadata(
        path=image, digest="Reader's", architectures=list(architectures)
    )



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Charmcraft Juju Config pydantic model."""
from typing import Annotated, Literal

import pydantic
from craft_application.models import CraftBaseModel


class _BaseJujuOption(CraftBaseModel):
    """A Juju option field. Do not use (use the child classes below)."""

    description: str | None = None
    default: str | int | float | bool | None = None


class JujuStringOption(_BaseJujuOption):
    """A Juju option field containing a string."""

    type: Literal["string"]
    default: str | None = None


class JujuIntOption(_BaseJujuOption):
    """A Juju option field containing an integer."""

    type: Literal["int"]
    default: pydantic.StrictInt | None = None


class JujuFloatOption(_BaseJujuOption):
    """A Juju option field containing a floating-point number."""

    type: Literal["float"]
    default: float | None = None


class JujuBooleanOption(_BaseJujuOption):
    """A Juju option field containing a boolean value."""

    type: Literal["boolean"]
    default: bool | None = None


class JujuSecretOption(_BaseJujuOption):
    """A Juju option field containing a secret ID."""

    type: Literal["secret"]
    # A secret doesn't really make sense, since it's unlikely
    # that anyone would know what the secret ID (specific to
    # the deployment in a model) is at the time that they are
    # writing the config, but included for completeness.
    default: (
        Annotated[str, pydantic.StringConstraints(pattern=r"^secret:[a-z0-9]{20}$")] | None
    ) = None


JujuOption = Annotated[
    JujuStringOption | JujuIntOption | JujuFloatOption | JujuBooleanOption | JujuSecretOption,
    pydantic.Field(discriminator="type"),
]


class JujuConfig(CraftBaseModel):
    """Juju configs for charms.

    See also: https://juju.is/docs/sdk/config
    and: https://juju.is/docs/sdk/config-yaml
    """

    options: dict[str, JujuOption] | None = None



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for the Charmcraft class."""

import pathlib

import pytest
from craft_application import util

from charmcraft import utils


@pytest.mark.parametrize(
    "charm_dir",
    [
        pytest.param(path, id=path.name)
        for path in sorted((pathlib.Path(__file__).parent / "sample-charms").iterdir())
    ],
)
def test_load_charm(app, charm_dir):
    app.project_dir = charm_dir

    project = app.get_project()
    with (charm_dir / "expected.yaml").open() as f:
        expected_data = util.safe_yaml_load(f)

    project_dict = project.marshal()

    assert project_dict == expected_data
    assert utils.dump_yaml(project_dict) == (charm_dir / "expected.yaml").read_text()



# Copyright 2020 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import io
import pathlib
import subprocess
import sys
import tempfile
import textwrap

import pytest

from charmcraft import jujuignore


def test_default_important_files():
    """Don't ignore important files by default."""
    ignore = jujuignore.JujuIgnore(jujuignore.default_juju_ignore)
    assert not ignore.match("version", is_dir=False)


def test_jujuignore_only_dir():
    ignore = jujuignore.JujuIgnore(["target/"])
    assert ignore.match("target", is_dir=True)
    assert ignore.match("foo/target", is_dir=True)
    assert not ignore.match("foo/1target", is_dir=True)
    assert not ignore.match("foo/target", is_dir=False)


def test_jujuignore_any_target():
    ignore = jujuignore.JujuIgnore(["target"])
    assert ignore.match("target", is_dir=True)
    assert ignore.match("/foo/target", is_dir=True)
    assert not ignore.match("/foo/1target", is_dir=True)
    assert ignore.match("/foo/target", is_dir=False)


def test_jujuignore_only_root_target_dir():
    ignore = jujuignore.JujuIgnore(["/target/"])
    assert ignore.match("/target", is_dir=True)
    assert not ignore.match("/foo/target", is_dir=True)
    assert not ignore.match("/foo/1target", is_dir=True)
    assert not ignore.match("/foo/target", is_dir=False)
    assert not ignore.match("/target", is_dir=False)


def test_jujuignore_only_root_target_file_or_dir():
    ignore = jujuignore.JujuIgnore(["/target"])
    assert ignore.match("/target", is_dir=True)
    assert not ignore.match("/foo/target", is_dir=True)
    assert not ignore.match("/foo/1target", is_dir=True)
    assert not ignore.match("/foo/target", is_dir=False)
    assert ignore.match("/target", is_dir=False)


def test_jujuignore_all_py_files():
    ignore = jujuignore.JujuIgnore(["*.py"])
    assert not ignore.match("/target", is_dir=True)
    assert ignore.match("/target.py", is_dir=True)
    assert ignore.match("/target.py", is_dir=False)
    assert ignore.match("/foo/target.py", is_dir=True)
    assert ignore.match("/foo/target.py", is_dir=False)
    assert not ignore.match("/target.pyT", is_dir=True)
    assert not ignore.match("/target.pyT", is_dir=False)


def test_jujuignore_all_py_files_in_foo():
    ignore = jujuignore.JujuIgnore(["/foo/*.py"])
    assert not ignore.match("/target.py", is_dir=True)
    assert not ignore.match("/target.py", is_dir=False)
    assert ignore.match("/foo/target.py", is_dir=True)
    assert ignore.match("/foo/target.py", is_dir=False)
    assert not ignore.match("/foo/sub/target.py", is_dir=False)


def test_jujuignore_ignore_comment():
    ignore = jujuignore.JujuIgnore(
        [
            "#comment",
            "target",
        ]
    )
    assert ignore.match("/target", is_dir=True)
    assert not ignore.match("/comment", is_dir=False)
    assert not ignore.match("/#comment", is_dir=False)
    assert not ignore.match("#comment", is_dir=False)
    assert not ignore.match("/foo/comment", is_dir=False)
    assert not ignore.match("/foo/#comment", is_dir=False)


def test_jujuignore_escaped_comment():
    ignore = jujuignore.JujuIgnore(
        [
            "\\#comment",
            "target",
        ]
    )
    assert ignore.match("/target", is_dir=True)
    assert ignore.match("/#comment", is_dir=False)
    assert ignore.match("#comment", is_dir=False)
    assert not ignore.match("/foo/comment", is_dir=False)
    assert ignore.match("/foo/#comment", is_dir=False)


def test_jujuignore_subdirectory_match():
    ignore = jujuignore.JujuIgnore(["apps/logs/"])
    assert not ignore.match("/logs", is_dir=True)
    assert not ignore.match("/apps", is_dir=True)
    assert ignore.match("/apps/logs", is_dir=True)
    assert not ignore.match("/apps/logs", is_dir=False)
    assert not ignore.match("/apps/foo/logs", is_dir=True)


def test_jujuignore_sub_subdirectory_match():
    ignore = jujuignore.JujuIgnore(["apps/*/logs/"])
    assert not ignore.match("/logs", is_dir=True)
    assert not ignore.match("/apps", is_dir=True)
    assert not ignore.match("/apps/logs", is_dir=True)
    assert not ignore.match("/apps/logs", is_dir=False)
    assert ignore.match("/apps/foo/logs", is_dir=True)
    assert ignore.match("/apps/bar/logs", is_dir=True)
    assert not ignore.match("/apps/baz/logs", is_dir=False)
    assert not ignore.match("/apps/foo/bar/logs", is_dir=True)


def test_jujuignore_any_subdirectory_match():
    ignore = jujuignore.JujuIgnore(["apps/**/logs/"])
    assert not ignore.match("/logs", is_dir=True)
    assert not ignore.match("/apps", is_dir=True)
    assert ignore.match("/apps/logs", is_dir=True)
    assert not ignore.match("/apps/logs", is_dir=False)
    assert ignore.match("/apps/foo/logs", is_dir=True)
    assert ignore.match("/apps/bar/logs", is_dir=True)
    assert ignore.match("/apps/bar/bing/logs", is_dir=True)
    assert not ignore.match("/apps/baz/logs", is_dir=False)


def test_jujuignore_everything_under_foo():
    ignore = jujuignore.JujuIgnore(["foo/**"])
    assert not ignore.match("/foo", is_dir=True)
    assert ignore.match("/foo/a", is_dir=True)
    assert ignore.match("/foo/a", is_dir=False)
    assert ignore.match("/foo/a/b", is_dir=False)


def test_jujuignore_everything_under_foo_but_readme():
    ignore = jujuignore.JujuIgnore(
        [
            "foo/**",
            "!foo/README.md",
        ]
    )
    assert not ignore.match("/foo", is_dir=True)
    assert ignore.match("/foo/a", is_dir=True)
    assert ignore.match("/foo/a", is_dir=False)
    assert ignore.match("/foo/a/b", is_dir=False)
    assert not ignore.match("/foo/README.md", is_dir=False)


def test_jujuignore_negation():
    ignore = jujuignore.JujuIgnore(
        [
            "*.py",
            "!foo.py",
            "!!!bar.py",
        ]
    )
    assert not ignore.match("bar.py", is_dir=False)
    assert not ignore.match("foo.py", is_dir=False)
    assert ignore.match("baz.py", is_dir=False)
    assert not ignore.match("foo/bar.py", is_dir=False)
    assert not ignore.match("foo/foo.py", is_dir=False)
    assert ignore.match("foo/baz.py", is_dir=False)


def test_jujuignore_multiple_doublestar():
    ignore = jujuignore.JujuIgnore(["foo/**/bar/**/baz"])
    assert ignore.match("/foo/1/2/bar/baz", is_dir=True)
    assert ignore.match("/foo/1/2/bar/1/2/baz", is_dir=True)
    assert ignore.match("/foo/bar/1/2/baz", is_dir=True)
    assert ignore.match("/foo/1/bar/baz", is_dir=True)
    assert ignore.match("/foo/bar/baz", is_dir=True)


def test_jujuignore_trim_unescaped_trailing_space():
    ignore = jujuignore.JujuIgnore(
        [
            r"foo  ",
            r"bar\ \ ",
        ]
    )
    assert ignore.match("/foo", is_dir=True)
    assert not ignore.match("/bar", is_dir=True)
    assert ignore.match("/bar  ", is_dir=True)


def test_jujuignore_escaped_bang():
    ignore = jujuignore.JujuIgnore(
        [
            r"foo",
            r"!/foo",
            r"\!foo",
        ]
    )
    assert not ignore.match("/foo", is_dir=True)
    assert ignore.match("/bar/foo", is_dir=True)
    assert ignore.match("!foo", is_dir=True)


def test_jujuignore_leading_whitespace():
    ignore = jujuignore.JujuIgnore(
        [
            r"foo",
            r" bar",
            r" #comment",
        ]
    )
    assert ignore.match("/foo", is_dir=True)
    assert ignore.match("/bar", is_dir=True)
    assert ignore.match("/bar/foo", is_dir=True)
    assert not ignore.match("#comment", is_dir=True)
    assert not ignore.match("comment", is_dir=True)


def test_jujuignore_bracket():
    ignore = jujuignore.JujuIgnore(
        [
            r"*.py[cod]",
        ]
    )
    assert not ignore.match("foo.py", is_dir=False)
    assert ignore.match("foo.pyc", is_dir=False)
    assert ignore.match("foo.pyo", is_dir=False)
    assert ignore.match("foo.pyd", is_dir=False)


def test_jujuignore_simple_match():
    ignore = jujuignore.JujuIgnore(
        [
            r"foo",
        ]
    )
    assert ignore.match("/foo", is_dir=True)
    assert ignore.match("/bar/foo", is_dir=True)
    assert not ignore.match("/bar/bfoo", is_dir=True)
    assert not ignore.match("/bfoo", is_dir=True)


def test_jujuignore_star_match():
    ignore = jujuignore.JujuIgnore(
        [
            r"foo*.py",
        ]
    )
    assert ignore.match("/foo.py", is_dir=False)
    assert ignore.match("/foo2.py", is_dir=False)
    assert ignore.match("/foobar.py", is_dir=False)
    assert ignore.match("/bar/foo.py", is_dir=False)
    assert not ignore.match("/foo/2.py", is_dir=False)


def test_jujuignore_paths_with_newlines():
    ignore = jujuignore.JujuIgnore([r"bar/**/*.py"])
    assert ignore.match("/bar/foo.py", is_dir=False)
    assert ignore.match("/bar/f\noo.py", is_dir=False)
    assert ignore.match("/bar/baz/f\noo.py", is_dir=False)
    assert ignore.match("/bar/b\nz/f\noo.py", is_dir=False)


def test_rstrip_unescaped():
    assert jujuignore._rstrip_unescaped(r"") == ""
    assert jujuignore._rstrip_unescaped(r" ") == ""
    assert jujuignore._rstrip_unescaped(r"a") == "a"
    assert jujuignore._rstrip_unescaped(r"a ") == "a"
    assert jujuignore._rstrip_unescaped(r"a  ") == "a"
    assert jujuignore._rstrip_unescaped(r"a\  ") == r"a\ "
    assert jujuignore._rstrip_unescaped(r"a foo\  ") == r"a foo\ "


def test_unescape_rule():
    assert jujuignore._unescape_rule(r"") == ""
    assert jujuignore._unescape_rule(r" ") == ""
    assert jujuignore._unescape_rule(r"\#") == "#"
    assert jujuignore._unescape_rule(r"\!") == "!"
    assert jujuignore._unescape_rule(r"\ ") == " "
    assert jujuignore._unescape_rule(r" foo\ ") == "foo "


def test_from_file():
    content = io.StringIO(
        textwrap.dedent(
            """\
    foo
    /bar
    """
        )
    )
    ignore = jujuignore.JujuIgnore(content)
    assert ignore.match("foo", is_dir=False)
    assert ignore.match("/foo", is_dir=False)
    assert ignore.match("/bar", is_dir=False)
    assert not ignore.match("/foo/bar", is_dir=False)


def test_log_matching_rule(assert_output):
    jujuignore.JujuIgnore(["foo/bar\n"])
    assert_output(r"Translated .jujuignore 1 'foo/bar\n' => '.*/foo/bar\\Z'")


def assert_matched_and_non_matched(globs, matched, unmatched, skip_git=False):
    """For a given set of globs, check that it does and doesn't match as expected"""
    ignore = jujuignore.JujuIgnore(globs)
    for m in matched:
        assert ignore.match(m, is_dir=False), f"{m} should have matched"
    for m in unmatched:
        assert not ignore.match(m, is_dir=False), f"{m} should not have matched"
    if skip_git:
        return
    with tempfile.TemporaryDirectory() as tmpdir:
        subprocess.run(["git", "init", tmpdir], check=True)
        with open(str(pathlib.Path(tmpdir) / ".gitignore"), "w") as gitignore:
            gitignore.writelines([g + "\n" for g in globs])
        input = "".join([m.lstrip("/") + "\n" for m in matched + unmatched])
        check = True
        if len(matched) == 0:
            # We don't check git return value because it returns nonzero if no paths match
            check = False
        p = subprocess.run(
            ["git", "check-ignore", "--stdin"],
            check=check,
            input=input,
            stdout=subprocess.PIPE,
            cwd=tmpdir,
            text=True,
        )
    matched_out = p.stdout.splitlines()
    assert sorted(matched) == sorted(
        matched_out
    ), f"expected exactly {matched} to match not {matched_out}"


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_star_vs_star_start():
    assert_matched_and_non_matched(
        ["/*.py", "**/foo"],
        # Only top level .py files, but foo at any level
        ["a.py", "b.py", "foo", "bar/foo"],
        ["bar/b.py"],
        # 'foo/a.py',  git matches foo/a.py because of foo, ours doesn't but I don't think it
        # matters because the whole directory would have already been skipped
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_questionmark():
    assert_matched_and_non_matched(
        ["foo?.py"],
        ["fooa.py", "foob.py"],
        ["foo.py", "footwo.py", "foo/.py"],
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_brackets():
    assert_matched_and_non_matched(
        ["*.py[cod]"],
        ["a.pyc", "b.pyo", "d.pyd", "foo/.pyc", "bar/__pycache__.pyc"],
        ["a.py", "b.pyq", "c.so", "foo/__pycache__/bar.py"],
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_bracket_ranges():
    assert_matched_and_non_matched(
        ["foo[1-9].py"],
        ["foo1.py", "foo2.py", "foo9.py"],
        ["foo0.py", "foo10.py", "fooa.py"],
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_bracket_inverted():
    assert_matched_and_non_matched(
        ["foo[!1-9].py", "bar[!a].py"],
        ["fooa.py", "foob.py", "fooc.py", "barb.py", "barc.py"],
        ["foo1.py", "foo2.py", "foo10.py", "bara.py"],
    )


def test_slashes_in_brackets():
    assert_matched_and_non_matched(
        [r"foo[\\].py"],
        [r"foo\.py"],
        [r"fooa.py"],
        # We don't test against git here, because it replies with "foo\\.py"
        # which is an escaped form that we'd have to interpret
        skip_git=True,
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_special_chars_in_brackets():
    assert_matched_and_non_matched(
        [r"foo[a|b].py"],
        [r"fooa.py", "foob.py", "foo|.py"],
        [r"foo.py", r"fooc.py"],
    )
    assert_matched_and_non_matched(
        [r"foo[ab|cd].py"],
        [r"fooa.py", "foob.py", "fooc.py", "food.py", "foo|.py"],
        [r"foo.py", "fooe.py", "fooab.py", "fooac.py"],
    )
    assert_matched_and_non_matched(
        [r"foo[a&].py"],
        [r"fooa.py", r"foo&.py"],
        [r"foo.py", r"fooa&.py", "foob.py"],
    )
    assert_matched_and_non_matched(
        [r"foo[a~].py"],
        [r"fooa.py", r"foo~.py"],
        [r"foo.py", r"fooa~.py", "foob.py"],
    )
    assert_matched_and_non_matched(
        [r"foo[[a].py"],
        [r"fooa.py", r"foo[.py"],
        [r"foo.py", r"fooa[.py", "foob.py"],
    )
    # Git allows ! or ^ to mean negate the glob
    assert_matched_and_non_matched(
        [r"foo[^a].py"],
        ["foob.py", "fooc.py", "foo^.py"],
        [r"foo.py", r"fooa.py", r"fooa^.py"],
    )


def test_extend_patterns():
    ignore = jujuignore.JujuIgnore(["foo"])
    assert ignore.match("foo", is_dir=False)
    assert not ignore.match("bar", is_dir=False)
    ignore.extend_patterns(["bar"])
    assert ignore.match("foo", is_dir=False)
    assert ignore.match("bar", is_dir=False)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for store commands."""
import argparse
import datetime
import pathlib
import textwrap
from unittest import mock

import craft_cli.pytest_plugin
import craft_store
import pytest
from craft_store import models

from charmcraft import errors, store
from charmcraft.application.commands import SetResourceArchitecturesCommand
from charmcraft.application.commands.store import FetchLibs, LoginCommand
from charmcraft.application.main import APP_METADATA
from charmcraft.models.project import CharmLib
from charmcraft.utils import cli
from tests import get_fake_revision

BASIC_CHARMCRAFT_YAML = """\
type: charm
"""


def test_login_basic_no_export(service_factory, mock_store_client):
    cmd = LoginCommand({"app": APP_METADATA, "services": service_factory})

    cmd.run(
        argparse.Namespace(
            charm=None,
            bundle=None,
            channel=None,
            permission=None,
            ttl=None,
            export=None,
        )
    )


@pytest.mark.parametrize("charm", [None, ["my-charm"]])
@pytest.mark.parametrize("bundle", [None, ["my-bundle"]])
@pytest.mark.parametrize("channel", [None, ["edge", "latest/stable"]])
@pytest.mark.parametrize("permission", [None, [], ["package-manage"]])
@pytest.mark.parametrize("ttl", [None, 0, 2**65])
def test_login_export(
    monkeypatch, service_factory, mock_store_client, charm, bundle, channel, permission, ttl
):
    mock_client_cls = mock.Mock(return_value=mock_store_client)
    monkeypatch.setattr(craft_store, "StoreClient", mock_client_cls)
    mock_store_client.login.return_value = "Some store credentials"
    cmd = LoginCommand({"app": APP_METADATA, "services": service_factory})

    cmd.run(
        argparse.Namespace(
            charm=charm,
            bundle=bundle,
            channel=channel,
            permission=permission,
            ttl=ttl,
            export=pathlib.Path("charmhub.login"),
        )
    )

    assert pathlib.Path("charmhub.login").read_text() == "Some store credentials"
    mock_store_client.login.assert_called_once()


@pytest.mark.parametrize(
    ("updates", "expected"),
    [
        ([], []),
        (
            [
                get_fake_revision(
                    revision=123,
                    updated_at=datetime.datetime(1900, 1, 1),
                    bases=[models.ResponseCharmResourceBase()],
                )
            ],
            [{"revision": 123, "updated_at": "1900-01-01T00:00:00", "architectures": ["all"]}],
        ),
    ],
)
def test_set_resource_architectures_output_json(emitter, updates, expected):
    SetResourceArchitecturesCommand.write_output(cli.OutputFormat.JSON, updates)

    emitter.assert_json_output(expected)


@pytest.mark.parametrize(
    ("updates", "expected"),
    [
        ([], "No revisions updated."),
        (
            [
                get_fake_revision(
                    revision=123,
                    updated_at=datetime.datetime(1900, 1, 1),
                    bases=[models.ResponseCharmResourceBase()],
                )
            ],
            textwrap.dedent(
                """\
                  Revision  Updated At            Architectures
                ----------  --------------------  ---------------
                       123  1900-01-01T00:00:00Z  all"""
            ),
        ),
    ],
)
def test_set_resource_architectures_output_table(emitter, updates, expected):
    SetResourceArchitecturesCommand.write_output(cli.OutputFormat.TABLE, updates)

    emitter.assert_message(expected)


def test_fetch_libs_no_charm_libs(
    emitter: craft_cli.pytest_plugin.RecordingEmitter, service_factory
):
    fetch_libs = FetchLibs({"app": APP_METADATA, "services": service_factory})

    with pytest.raises(errors.LibraryError) as exc_info:
        fetch_libs.run(argparse.Namespace())

    assert exc_info.value.resolution == "Add a 'charm-libs' section to charmcraft.yaml."


@pytest.mark.parametrize(
    ("libs", "expected"),
    [
        (
            [CharmLib(lib="mysql.mysql", version="1")],
            textwrap.dedent(
                """\
                Could not find the following libraries on charmhub:
                - lib: mysql.mysql
                  version: '1'
                """
            ),
        ),
        (
            [
                CharmLib(lib="mysql.mysql", version="1"),
                CharmLib(lib="some_charm.lib", version="1.2"),
            ],
            textwrap.dedent(
                """\
                Could not find the following libraries on charmhub:
                - lib: mysql.mysql
                  version: '1'
                - lib: some-charm.lib
                  version: '1.2'
                """
            ),
        ),
    ],
)
def test_fetch_libs_missing_from_store(service_factory, libs, expected):
    service_factory.project.charm_libs = libs
    service_factory.store.anonymous_client.fetch_libraries_metadata.return_value = []
    fetch_libs = FetchLibs({"app": APP_METADATA, "services": service_factory})

    with pytest.raises(errors.CraftError) as exc_info:
        fetch_libs.run(argparse.Namespace())

    assert exc_info.value.args[0] == expected


@pytest.mark.parametrize(
    ("libs", "store_libs", "dl_lib", "expected"),
    [
        (
            [CharmLib(lib="mysql.backups", version="1")],
            [
                store.models.Library(
                    charm_name="mysql",
                    lib_name="backups",
                    lib_id="ididid",
                    api=1,
                    patch=2,
                    content=None,
                    content_hash="hashhashhash",
                )
            ],
            store.models.Library(
                charm_name="mysql",
                lib_name="backups",
                lib_id="ididid",
                api=1,
                patch=2,
                content=None,
                content_hash="hashhashhash",
            ),
            "Store returned no content for 'mysql.backups'",
        ),
    ],
)
def test_fetch_libs_no_content(new_path, service_factory, libs, store_libs, dl_lib, expected):
    service_factory.project.charm_libs = libs
    service_factory.store.anonymous_client.fetch_libraries_metadata.return_value = store_libs
    service_factory.store.anonymous_client.get_library.return_value = dl_lib
    fetch_libs = FetchLibs({"app": APP_METADATA, "services": service_factory})

    with pytest.raises(errors.CraftError, match=expected) as exc_info:
        fetch_libs.run(argparse.Namespace())

    assert exc_info.value.args[0] == expected


@pytest.mark.parametrize(
    ("libs", "store_libs", "dl_lib", "expected"),
    [
        (
            [CharmLib(lib="mysql.backups", version="1")],
            [
                store.models.Library(
                    charm_name="mysql",
                    lib_name="backups",
                    lib_id="ididid",
                    api=1,
                    patch=2,
                    content=None,
                    content_hash="hashhashhash",
                )
            ],
            store.models.Library(
                charm_name="mysql",
                lib_name="backups",
                lib_id="ididid",
                api=1,
                patch=2,
                content="I am a library.",
                content_hash="hashhashhash",
            ),
            "Store returned no content for 'mysql.backups'",
        ),
    ],
)
def test_fetch_libs_success(
    new_path, emitter, service_factory, libs, store_libs, dl_lib, expected
) -> None:
    service_factory.project.charm_libs = libs
    service_factory.store.anonymous_client.fetch_libraries_metadata.return_value = store_libs
    service_factory.store.anonymous_client.get_library.return_value = dl_lib
    fetch_libs = FetchLibs({"app": APP_METADATA, "services": service_factory})

    fetch_libs.run(argparse.Namespace())

    emitter.assert_progress("Getting library metadata from charmhub")
    emitter.assert_message("Downloaded 1 charm libraries.")
